;*************************************************************************
; Title	:    I2C Slave Implementation (based on Atmel Appl. Note AVR302)
;
; Author:    DG4KLU
; Target:    any AVR device
;*************************************************************************

#include <avr/io.h>

#define IO_SREG  _SFR_IO_ADDR(SREG)
#define IO_MCUCR _SFR_IO_ADDR(MCUCR)
#define IO_TIMSK _SFR_IO_ADDR(TIMSK)
#define IO_GIMSK _SFR_IO_ADDR(GIMSK)
#define IO_TCCR0 _SFR_IO_ADDR(TCCR0)
#define IO_TCNT0 _SFR_IO_ADDR(TCNT0)
#define IO_DDRD  _SFR_IO_ADDR(DDRD)
#define IO_PORTD _SFR_IO_ADDR(PORTD)
#define IO_PIND  _SFR_IO_ADDR(PIND)
#define IO_PORTB _SFR_IO_ADDR(PORTB)
#define IO_DDRB  _SFR_IO_ADDR(DDRB)

;**** Global I2C Constants ****

	devadr	= 0x50		; Slave device address
	devadrm	= 0x7F		; Slave multi address mask

	pinmask = 0x14		; <=> (1<<PD4)+(1<<PD2)

;**** Global Register Variables ****

	temp	= 16		; Temporary variable
	etemp	= 17		; Extra temporary variable
	I2Cdata	= 18		; I2C data register
	I2Cadr	= 19		; I2C address and direction register
	I2Cstat	= 20		; I2C temporary SREG storage

;***************************************************************************
;*
;* INTERRUPT
;*	EXT_INT0 / I2C_wakeup
;*
;* DESCRIPTION
;*	Detects the start condition and handles the data transfer on 
;*	the I2C bus.
;*
;*	Received data is stored in the r0 register and r0 is transmitted 
;*	during a master read transfer.
;*
;* NOTE
;*	This interrupt code, in some cases, jumps to code parts in the 
;*	TIM0_OVF / I2C_skip section to reduce code size.
;*
;***************************************************************************

.global SIG_INTERRUPT0
SIG_INTERRUPT0:			; INT0 handle
	in	I2Cstat,IO_SREG	; store SREG


;*************************
;* Get I2C Slave Address *
;*************************

I2C_get_adr:
	ldi	I2Cadr,1	; initialize address register
wlo_ga0:sbic	IO_PIND,PIND4	; 	wait for SCL low
	rjmp	wlo_ga0
	rjmp	first_ga
do_ga:				; do
wlo_ga:	sbic	IO_PIND,PIND4	; 	wait for SCL low
	rjmp	wlo_ga

	mov	temp,I2Cadr	;	test address
	andi	temp,devadrm	; 	mask "register full" floating bit
	cpi	temp,devadr	; 	if device addressed
	in	temp,IO_SREG	;		set T flag ("T = Z")
	bst	temp,1

first_ga:
	sec			; 	set carry
whi_ga:	sbis	IO_PIND,PIND4	; 	wait for SCL high
	rjmp	whi_ga
	sbis	IO_PIND,PIND2	; 	if SDA low
	clc			;		clear carry
	rol	I2Cadr		; 	shift carry into address register

	brcc	do_ga		; while register not full

wlo_ca:	sbic	IO_PIND,PIND4	; wait for SCL low
	rjmp	wlo_ca

;**** Check Address ****
				; if T flag set (device addressed)
	brts	I2C_adr_ack	;	acknowledge address
				; else
	rjmp	I2C_adr_miss	; 	goto address miss handle

;**** Acknowledge Address ****

I2C_adr_ack:
	sbi	IO_DDRD,DDD2	; assert acknowledge on SDA
whi_aa:	sbis	IO_PIND,PIND4	; wait for SCL high
	rjmp	whi_aa

;**** Check Transfer Direction ****

	lsr	I2Cadr		; if master write
	brcc	I2C_master_write;	goto master write handle	


;*******************************
;* Transmit Data (master read) *
;*******************************

I2C_master_read:

;**** Load Data (handle outgoing data) ****

;	sbi	IO_DDRD,DDD4	; (start wait state)

	in	I2Cdata,IO_PORTB	; read current state of PORTB

;	cbi	IO_DDRD,DDD4	; (end wait state)

	sec			; shift MSB out and "floating empty flag" in
	rol	I2Cdata

wlo_mr:	sbic	IO_PIND,PIND4	; wait for SCL low
	rjmp	wlo_mr


;**** Transmitt data ****

	brcc	fb_low_mr	; if current data bit high
	cbi	IO_DDRD,DDD2	;	release SDA
	rjmp	fb_mr		;	goto read loop
fb_low_mr:			; else
	sbi	IO_DDRD,DDD2	;	force SDA
fb_mr:	lsl	I2Cdata		; 	if data register not empty

loop_mr:
whi_mr:	sbis	IO_PIND,PIND4	; wait for SCL high
	rjmp	whi_mr
wlo_mr2:sbic	IO_PIND,PIND4	; wait for SCL low
	rjmp	wlo_mr2

	brcc	b_low_mr	; if current data bit high

	cbi	IO_DDRD,DDD2	;	release SDA
	lsl	I2Cdata		; 	if data register not empty
	brne	loop_mr		;		loop
	rjmp	done_mr		;	done
b_low_mr:			; else
	sbi	IO_DDRD,DDD2	;	force SDA
	lsl	I2Cdata		; 	if data register not empty
	brne	loop_mr		;		loop

done_mr:
whi_mr2:sbis	IO_PIND,PIND4	; wait for SCL high
	rjmp	whi_mr2
wlo_mr3:sbic	IO_PIND,PIND4	; wait for SCL low
	rjmp	wlo_mr3

	cbi	IO_DDRD,DDD2	; release SDA

;**** Read Acknowledge from Master ****

whi_ra:	sbis	IO_PIND,PIND4	; wait for SCL high
	rjmp	whi_ra

	sec			; read acknowledge
	sbis	IO_PIND,PIND2
	clc
	brcc	I2C_master_read	; if ack transfer (next) data

wlo_ra:	sbic	IO_PIND,PIND4	; wait for SCL low
	rjmp	wlo_ra

	rjmp	I2C_wait_cond	; goto wait condition (rep. start or stop)


;*******************************
;* Receive Data (master write) *
;*******************************

I2C_master_write:

wlo_mw0:sbic	IO_PIND,PIND4	; wait for SCL low
	rjmp	wlo_mw0
	cbi	IO_DDRD,DDD2	; remove acknowledge from SDA

whi_mw:	sbis	IO_PIND,PIND4	; wait for SCL high
	rjmp	whi_mw

	in	temp,IO_PIND	; sample SDA (first bit) and SCL
	andi	temp,pinmask	; mask out SDA and SCL
do_mw:				; do
	in	etemp,IO_PIND	;	new sample
	andi	etemp,pinmask	;	mask out SDA and SCL
	cp	etemp,temp
	breq	do_mw		; while no change

	sbrs	etemp,PIND4	; if SCL changed to low
	rjmp	receive_data	;	goto receive data

	sbrs	etemp,PIND2	; if SDA changed to low
	rjmp	I2C_get_adr	;	goto repeated start
				; else
	rjmp	I2C_stop	;	goto transfer stop

receive_data:
	ldi	I2Cdata,2	; set I2Cdata MSB to zero
	sbrc	temp,PIND2	; if SDA sample is one
	ldi	I2Cdata,3	;	set I2Cdata MSB to one

do_rd:				; do
wlo_rd:	sbic	IO_PIND,PIND4	;	wait for SCL low
	rjmp	wlo_rd
	sec			; 	set carry
whi_rd:	sbis	IO_PIND,PIND4	;	wait for SCL high
	rjmp	whi_rd

	sbis	IO_PIND,PIND2	; 	if SDA low
	clc			;		clear carry
	rol	I2Cdata		; 	shift carry into data register

	brcc	do_rd		; while register not full

;**** Acknowledge Data ****

I2C_dat_ack:
wlo_da:	sbic	IO_PIND,PIND4	; wait for SCL low
	rjmp	wlo_da
	sbi	IO_DDRD,DDD2	; assert acknowledge on SDA
whi_da:	sbis	IO_PIND,PIND4	; wait for SCL high
	rjmp	whi_da
				; (acknowledge is removed later)

;**** Store Data (handle incoming data) ****

;	sbi	IO_DDRD,DDD4	; (start wait state)

	out	IO_PORTB,I2Cdata	; write data to PORTB

;	cbi	IO_DDRD,DDD4	; (end wait state)

	rjmp	I2C_master_write; Start on next transfer


;***************************************************************************
;*
;* INTERRUPT
;*	TIM0_OVF / I2C_skip
;*
;* DESCRIPTION
;*	This interrupt handles a "address miss". If the slave device is
;*	not addressed by a master, it will not acknowledge the address.
;*
;*	Instead of waiting for a new start condition in a "busy loop"
;*	the slave set up the counter to count 8 falling edges on SCL and 
;*	returns from the current interrupt. This "skipping" of data
;*	do not occupies any processor time. When 8 egdes are counted, a
;*	timer overflow interrupt handling routine (this one) will check
;*	the next condition that accure. If it's a stop condition
;*	the transfer ends, if it's a repeated start condition a jump
;*	to the I2C_wakeup interrupt will read the new address. If a new
;*	transfer is initiated the "skipping" process is repeated.
;*
;***************************************************************************

.global SIG_OVERFLOW0
SIG_OVERFLOW0:			; ( Timer 0 overflow handle )
	in	I2Cstat,IO_SREG	; store SREG

I2C_adr_miss:
	
;**** Drop Acknowledge ****

whi_dac:sbis	IO_PIND,PIND4	; wait for SCL high
	rjmp	whi_dac
wlo_dac:sbic	IO_PIND,PIND4	; wait for SCL low
	rjmp	wlo_dac

	; disable timer 0 overflow interrupt
	ldi	temp,(0<<TOIE0)
	out	IO_TIMSK,temp
	; enable external interrupt request 0
	ldi	temp,(1<<INT0)
	out	IO_GIMSK,temp


;************************
;* Wait for a Condition *
;************************

I2C_wait_cond:
whi_wc:	sbis	IO_PIND,PIND4	; wait for SCL high
	rjmp	whi_wc

	in	temp,IO_PIND	; sample SDA (first bit) and SCL
	andi	temp,pinmask	; mask out SDA and SCL
do_wc:				; do
	in	etemp,IO_PIND	;	new sample
	andi	etemp,pinmask	;	mask out SDA and SCL
	cp	etemp,temp
	breq	do_wc		; while no change

	sbrs	etemp,PIND4	; if SCL changed to low
	rjmp	I2C_skip_byte	;	goto skip byte

	sbrs	etemp,PIND2	; if SDA changed to low
	rjmp	I2C_get_adr	;	goto repeated start
				; else
				;	goto transfer stop


;*************************
;* Handle Stop Condition *
;*************************

I2C_stop:
	;! Set INT0 to generate an interrupt on low level,
	;! then set INT0 to generate an interrupt on falling edge.
	;! This will clear the EXT_INT0 request flag.
	ldi	temp,(0<<ISC01)+(0<<ISC00)
	out	IO_MCUCR,temp
	ldi	temp,(1<<ISC01)+(0<<ISC00)
	out	IO_MCUCR,temp

	out	IO_SREG,I2Cstat	; restore SREG
	reti			; return from interrupt


;****************
;* Skip byte(s) *
;****************

I2C_skip_byte:
	; set counter initial value
	ldi	temp,-7
	out	IO_TCNT0,temp
	; enable timer 0 overflow interrupt
	ldi	temp,(1<<TOIE0)
	out	IO_TIMSK,temp
	; disable external interrupt request 0
	ldi	temp,(0<<INT0)
	out	IO_GIMSK,temp

	out	IO_SREG,I2Cstat	; restore SREG
	reti


;***************************************************************************
;*
;* FUNCTION
;*	I2C_init
;*
;* DESCRIPTION
;*	Initialization of interrupts and port used by the I2C interface
;*	and waits for the first start condition.
;*
;* USAGE
;*	Jump to this code directly after a reset.
;*
;* RETURN
;*	none
;*
;* NOTE
;*	Code size can be reduced by 12 instructions (words) if no transfer
;*	is started on the I2C bus (bus free), before the interrupt
;*	initialization is finished. If this can be ensured, remove the
;*	"Wait for I2C Start Condition" part and replace it with a "sei"
;*	instruction.
;*
;***************************************************************************

        .global I2C_init
        .func I2C_init
I2C_init:

;**** PORT Initialization ****

	; Initialize PD2 (INT0) for open colector operation (SDA in/out)
	ldi	temp,(0<<DDD4)+(0<<DDD2)
	out	IO_DDRD,temp

	; Initialize PD4 (T0) for open colector operation (SCL in/out)
	ldi	temp,(0<<PD4)+(0<<PD2)
	out	IO_PORTD,temp

	; Set PORTB to output an initialize all lines to 1
	ldi	temp,0xFF
	out	IO_DDRB,temp
	out	IO_PORTB,temp

;**** Interrupt Initialization ****

	; Set INT0 to generate an interrupt on falling edge
	ldi	temp,(1<<ISC01)+(0<<ISC00)
	out	IO_MCUCR,temp

	; Enable INT0
	ldi	temp,(1<<INT0)
	out	IO_GIMSK,temp

	; Set clock to count on falling edge of T0
	ldi	temp,(1<<CS02)+(1<<CS01)+(0<<CS00)
	out	IO_TCCR0,temp


;***************************************************
;* Wait for I2C Start Condition (13 intstructions) *
;***************************************************

do_start_w:			; do
	in	temp,IO_PIND	; 	sample SDA & SCL
	com	temp		;	invert
	andi	temp,pinmask	;	mask SDA and SCL
	brne	do_start_w	; while (!(SDA && SCL))

	in	temp,IO_PIND	; sample SDA & SCL
	andi	temp,pinmask	; mask out SDA and SCL
do_start_w2:			; do
	in	etemp,IO_PIND	;	new sample
	andi	etemp,pinmask	;	mask out SDA and SCL
	cp	etemp,temp
	breq	do_start_w2	; while no change

	sbrc	etemp,PIND2	; if SDA no changed to low
	rjmp	do_start_w	;	repeat

	rcall	I2C_get_adr	; call(!) interrupt handle (New transfer)
        
        ret
        .endfunc
        